/*
  Copyright (c) 2014, Matthias Schiffer <mschiffer@universe-factory.net>
  All rights reserved.

  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions are met:

    1. Redistributions of source code must retain the above copyright notice,
       this list of conditions and the following disclaimer.
    2. Redistributions in binary form must reproduce the above copyright notice,
       this list of conditions and the following disclaimer in the documentation
       and/or other materials provided with the distribution.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/


#pragma once

#include <cstdint>
#include <memory>
#include <stdexcept>
#include <unordered_map>
#include <vector>

#include "CollisionType.hpp"
#include "Entity.hpp"


namespace RPGEdit {

namespace Model {

class _Map {
protected:
	std::vector<std::string> tileset;

	size_t width, height;
	std::vector<CollisionType> collision;
	std::vector<std::vector<uint32_t>> tiles;

	_Map(size_t width0, size_t height0, size_t layers)
		: width(width0), height(height0), collision(width*height) {
		for (size_t i = 0; i < layers; i++)
			tiles.emplace_back(width*height);
	}
};

class Map : private _Map {
private:
	struct EntityState {
		Position position;

		Direction direction;
		uint64_t transitionStart = 0;
		uint64_t transitionEnd = 0;

		EntityState(const Position &position0) : position(position0) {}
	};

	std::vector<std::unique_ptr<Entity>> entities;
	std::unordered_map<const Entity *, EntityState> entityStates;


	void pushEntity(Entity *entity, EntityState &&state) {
		entities.push_back(std::unique_ptr<Entity>(entity));
		entityStates.emplace(entity, state);
	}

	void copyEntities(const Map &other) {
		for (auto &e : other.entities)
			pushEntity(new Entity(*e), EntityState(other.entityStates.at(e.get())));
	}


	Map(size_t width0, size_t height0, size_t layers) : _Map(width0, height0, layers) {
	}

public:
	Map & operator=(const Map &other) {
		static_cast<_Map>(*this) = other;
		copyEntities(other);

		return *this;
	}

	Map(const Map &other) : _Map(other) {
		copyEntities(other);
	}

	Map & operator=(Map &&other) {
		static_cast<_Map>(*this) = std::move(other);

		entities = std::move(other.entities);
		entityStates = std::move(other.entityStates);

		return *this;
	}

	Map(Map &&other) : _Map(std::move(other)) {
		entities = std::move(other.entities);
		entityStates = std::move(other.entityStates);
	}

	std::vector<std::string> & getTileset() {
		return tileset;
	}

	const std::vector<std::string> & getTileset() const {
		return tileset;
	}

	const std::vector<std::unique_ptr<Entity>> & getEntities() const {
		return entities;
	}

	Entity * addEntity(const std::string &name, const Position &pos) {
		Entity *e = new Entity(name);
		pushEntity(e, EntityState(pos));
		setCollisionAt(pos.x, pos.y, BLOCKED);
		return e;
	}

	size_t getWidth() const {
		return width;
	}

	size_t getHeight() const {
		return height;
	}

	size_t getLayerCount() const {
		return tiles.size();
	}

	void setCollisionAt(size_t x, size_t y, CollisionType value) {
		if (x >= width || y >= height)
			throw std::range_error("Map::setCollisionAt: bad coordinates");

		collision[y*width + x] = value;
	}

	CollisionType getCollisionAt(ssize_t x, ssize_t y) const {
		if (x < 0 || (size_t)x >= width || y < 0 || (size_t)y >= height)
			return BLOCKED;

		return collision[y*width + x];
	}


	void setTileAt(size_t layer, size_t x, size_t y, uint32_t value) {
		if (layer >= tiles.size() || x >= width || y >= height)
			throw std::range_error("Map::setTileAt: bad coordinates");

		tiles[layer][y*width + x] = value;
	}

	uint32_t getTileAt(size_t layer, ssize_t x, ssize_t y) const {
		if (layer >= tiles.size() || x < 0 || (size_t)x >= width || y < 0 || (size_t)y >= height)
			return 0;

		return tiles[layer][y*width + x];
	}

	Position getEntityPosition(const Entity *entity, uint64_t time) const;
	bool moveEntity(Entity *entity, Direction dir, uint64_t start, uint64_t end);
	void moveEntityTo(Entity *entity, Position pos);
	void finishEntityTransition(Entity *entity);

	static std::unique_ptr<Map> load(const std::string &name);
};

}

}
